import os
import unittest

import pytest
from torch_tb_profiler.profiler.data import RunProfileData
from torch_tb_profiler.profiler.diffrun import (compare_op_tree, diff_summary,
                                                print_node, print_ops)
from torch_tb_profiler.utils import timing


def load_profile(worker, span, path):
    return RunProfileData.parse(worker, span, path, '.')


class TestDiffRun(unittest.TestCase):

    @pytest.mark.skipif(not (os.path.isfile(os.path.expanduser('~/profile_result/worker0.pt.trace.json')) and
                        os.path.isfile(os.path.expanduser('~/profile_result/worker1.pt.trace.json'))),
                        reason="file doesn't exist")
    def test_happy_path(self):
        # path1 = os.path.expanduser('~/profile_result/worker0.pt.trace.json')
        path1 = '/home/mike/git/kineto/tb_plugin/examples/result/datapipe0.1638760942588.pt.trace.json'
        profile1 = load_profile('worker0', 1, path1)
        roots = list(profile1.tid2tree.values())
        root = roots[0]

        # path2 = os.path.expanduser('~/profile_result/worker1.pt.trace.json')
        path2 = '/home/mike/git/kineto/tb_plugin/examples/result/datapipe0.1638835897553.pt.trace.json'
        profile2 = load_profile('worker0', 1, path2)
        roots1 = list(profile2.tid2tree.values())
        root1 = roots1[0]

        with timing('Compare operator tree', True):
            node = compare_op_tree(root, root1)

        print_ops(node.children[4].left, prefix='    ')
        print('========================================================')
        print_ops(node.children[4].right)

        print('*********************** summary *************************')
        with timing('Diff summary', True):
            stats = diff_summary(node)

        # result = stats.flatten_diff_tree()
        # path = '0-1-1'
        # json_data = result[path].get_diff_node_summary(path)
        print_node(stats, 0, 0)


if __name__ == '__main__':
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
import logging
import math
import os
import time
from contextlib import contextmanager
from math import pow

from . import consts


def get_logging_level():
    log_level = os.environ.get('TORCH_PROFILER_LOG_LEVEL', 'INFO').upper()
    if log_level not in logging._levelToName.values():
        log_level = logging.getLevelName(logging.INFO)
    return log_level


logger = None


def get_logger():
    global logger
    if logger is None:
        logger = logging.getLogger(consts.PLUGIN_NAME)
        logger.setLevel(get_logging_level())
    return logger


def is_chrome_trace_file(path):
    return consts.WORKER_PATTERN.match(path)


def href(text, url):
    """"return html formatted hyperlink string

    Note:
        target="_blank" causes this link to be opened in new tab if clicked.
    """
    return f'<a href="{url}" target="_blank">{text}</a>'


class Canonicalizer:
    def __init__(
            self,
            time_metric='us',
            memory_metric='B',
            *,
            input_time_metric='us',
            input_memory_metric='B'):
        # raw timestamp is in microsecond
        # https://github.com/pytorch/pytorch/blob/v1.9.0/torch/csrc/autograd/profiler_kineto.cpp#L33
        time_metric_to_factor = {
            'us': 1,
            'ms': 1e3,
            's':  1e6,
        }
        # raw memory is in bytes
        memory_metric_to_factor = {
            'B':  pow(1024, 0),
            'KB': pow(1024, 1),
            'MB': pow(1024, 2),
            'GB': pow(1024, 3),
        }

        # canonicalize the memory metric to a string
        self.canonical_time_metrics = {
            'micro': 'us', 'microsecond': 'us', 'us': 'us',
            'milli': 'ms', 'millisecond': 'ms', 'ms': 'ms',
            '':  's',      'second':  's',  's':  's',
        }
        # canonicalize the memory metric to a string
        self.canonical_memory_metrics = {
            '':  'B',  'B':  'B',
            'K': 'KB', 'KB': 'KB',
            'M': 'MB', 'MB': 'MB',
            'G': 'GB', 'GB': 'GB',
        }

        self.time_metric = self.canonical_time_metrics[time_metric]
        self.memory_metric = self.canonical_memory_metrics[memory_metric]

        # scale factor scale input to output
        self.time_factor = time_metric_to_factor[self.canonical_time_metrics[input_time_metric]] /\
            time_metric_to_factor[self.time_metric]
        self.memory_factor = memory_metric_to_factor[self.canonical_memory_metrics[input_memory_metric]] /\
            memory_metric_to_factor[self.memory_metric]

    def convert_time(self, t):
        return self.time_factor * t

    def convert_memory(self, m):
        return self.memory_factor * m


class DisplayRounder:
    """Round a value for display purpose."""

    def __init__(self, ndigits):
        self.ndigits = ndigits
        self.precision = pow(10, -ndigits)

    def __call__(self, v: float):
        _v = abs(v)
        if _v >= self.precision or v == 0:
            return round(v, 2)
        else:
            ndigit = abs(math.floor(math.log10(_v)))
            return round(v, ndigit)


@contextmanager
def timing(description: str, force: bool = False) -> None:
    if force or os.environ.get('TORCH_PROFILER_BENCHMARK', '0') == '1':
        start = time.time()
        yield
        elapsed_time = time.time() - start
        logger.info(f'{description}: {elapsed_time}')
    else:
        yield



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
import re
from collections import namedtuple

PLUGIN_NAME = 'pytorch_profiler'

WORKER_PATTERN = re.compile(r"""^(.*?) # worker name
        (\.\d+)? # optional timestamp like 1619499959628 used as span name
        \.pt\.trace\.json # the ending suffix
        (?:\.gz)?$""", re.X)  # optional .gz extension

NODE_PROCESS_PATTERN = re.compile(r"""^(.*)_(\d+)""")
MONITOR_RUN_REFRESH_INTERNAL_IN_SECONDS = 10
MAX_GPU_PER_NODE = 64

View = namedtuple('View', 'id, name, display_name')
OVERALL_VIEW = View(1, 'overall', 'Overview')
OP_VIEW = View(2, 'operator', 'Operator')
KERNEL_VIEW = View(3, 'kernel', 'Kernel')
TRACE_VIEW = View(4, 'trace', 'Trace')
DISTRIBUTED_VIEW = View(5, 'distributed', 'Distributed')
MEMORY_VIEW = View(6, 'memory', 'Memory')
MODULE_VIEW = View(7, 'module', 'Module')
LIGHTNING_VIEW = View(8, 'lightning', 'Lightning')

TOOLTIP_GPU_UTIL = \
    'GPU Utilization:\n' \
    'GPU busy time / All steps time. The higher, the better. ' \
    'GPU busy time is the time during which there is at least one GPU kernel running on it. ' \
    'All steps time is the total time of all profiler steps(or called as iterations).\n'
TOOLTIP_SM_EFFICIENCY = \
    'Est. SM Efficiency:\n' \
    'Estimated Stream Multiprocessor Efficiency. The higher, the better. ' \
    'This metric of a kernel, SM_Eff_K = min(blocks of this kernel / SM number of this GPU, 100%). ' \
    "This overall number is the sum of all kernels' SM_Eff_K weighted by kernel's execution duration, " \
    'divided by all steps time.\n'
TOOLTIP_OCCUPANCY_COMMON = \
    'Est. Achieved Occupancy:\n' \
    'For most cases such as memory bandwidth bounded kernels, the higher the better. ' \
    'Occupancy is the ratio of active warps on an SM ' \
    'to the maximum number of active warps supported by the SM. ' \
    'The theoretical occupancy of a kernel is upper limit occupancy of this kernel, ' \
    'limited by multiple factors such as kernel shape, kernel used resource, ' \
    'and the GPU compute capability.\n' \
    'Est. Achieved Occupancy of a kernel, OCC_K = ' \
    'min(threads of the kernel / SM number / max threads per SM, theoretical occupancy of the kernel). '
TOOLTIP_OCCUPANCY_OVERVIEW = \
    "This overall number is the weighted average of all kernels' OCC_K " \
    "using kernel's execution duration as weight. " \
    'It shows fine-grained low-level GPU utilization.\n'
TOOLTIP_TENSOR_CORES = \
    'Kernel using Tensor Cores:\n' \
    'Total GPU Time for Tensor Core kernels / Total GPU Time for all kernels.\n'
TOOLTIP_OCCUPANCY_TABLE = \
    "This \"Mean\" number is the weighted average of all calls' OCC_K of the kernel, " \
    "using each call's execution duration as weight. " \
    'It shows fine-grained low-level GPU utilization.'
TOOLTIP_BLOCKS_PER_SM = \
    'Blocks Per SM = blocks of this kernel / SM number of this GPU.\n' \
    'If this number is less than 1, it indicates the GPU multiprocessors are not fully utilized.\n' \
    '\"Mean Blocks per SM\" is the weighted average of all calls of this kernel, ' \
    "using each call's execution duration as weight."
TOOLTIP_OP_TC_ELIGIBLE = \
    'Whether this operator is eligible to use Tensor Cores.'
TOOLTIP_OP_TC_SELF = \
    'Time of self-kernels with Tensor Cores / Time of self-kernels.'
TOOLTIP_OP_TC_TOTAL = \
    'Time of kernels with Tensor Cores / Time of kernels.'
TOOLTIP_KERNEL_USES_TC = \
    'Whether this kernel uses Tensor Cores.'
TOOLTIP_KERNEL_OP_TC_ELIGIBLE = \
    'Whether the operator launched this kernel is eligible to use Tensor Cores.'



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
from typing import Dict, Iterable, List

from .. import consts
from .node import OperatorNode


class TensorCoresParser:
    def __init__(self, tc_ratio: List[float], tc_eligible_ops_kernel_ratio: float):
        # For calculating Tensor Cores time ratio per GPU.
        self.tc_ratio = tc_ratio
        self.tc_eligible_ops_kernel_ratio = tc_eligible_ops_kernel_ratio

    @classmethod
    def parse_events(cls, tid2tree: Dict[str, OperatorNode], ops: Iterable[OperatorNode], gpu_ids: Iterable[int]):
        tc_ratio = cls._calculate_tc_ratio(ops, gpu_ids)
        tc_eligible_ops_kernel_ratio = cls._get_tc_eligible_ops_kernel_ratio(tid2tree, ops)
        return cls(tc_ratio, tc_eligible_ops_kernel_ratio)

    @staticmethod
    def _calculate_tc_ratio(ops: Iterable[OperatorNode], gpu_ids: Iterable[int]):
        tc_ratio: List[float] = [None] * consts.MAX_GPU_PER_NODE
        tc_time = [0] * consts.MAX_GPU_PER_NODE
        total_time = [0] * consts.MAX_GPU_PER_NODE
        has_kernel = False
        for op in ops:
            for rt in op.runtimes:
                # 'CallTreeRoot' & 'dummy' kernels are launched out of profiler step, so don't count them.
                if not (op.name == 'CallTreeRoot' and rt.name == 'dummy'):
                    for k in rt.get_kernels():
                        has_kernel = True
                        dur = k.end_time - k.start_time
                        is_tc_used = k.tc_used
                        if is_tc_used:
                            tc_time[k.device_id] += dur
                        total_time[k.device_id] += dur
        if has_kernel:  # If no kernel, then keep all self.tc_ratio as None.
            for gpu_id in gpu_ids:
                if total_time[gpu_id] > 0:
                    tc_ratio[gpu_id] = tc_time[gpu_id] / total_time[gpu_id]
                else:
                    tc_ratio[gpu_id] = 0.0
        return tc_ratio

    @staticmethod
    def _get_bottom_tc_eligible_operators(op_tree_node: OperatorNode):
        ops: List[OperatorNode] = []
        for child in op_tree_node.children:
            child_ops = TensorCoresParser._get_bottom_tc_eligible_operators(child)
            ops.extend(child_ops)
        # TC-eligible ops which have children TC-eligible ops will not be regarded as 'bottom'.
        if op_tree_node.tc_eligible and len(ops) == 0:
            ops.append(op_tree_node)
        return ops

    @staticmethod
    def _get_tc_eligible_ops_kernel_ratio(tid2tree: Dict[int, OperatorNode], ops: Iterable[OperatorNode]):
        def sum_self_kernel_time(ops: Iterable[OperatorNode]):
            sum_time = 0
            for op in ops:
                for rt in op.runtimes:
                    # 'CallTreeRoot' & 'dummy' kernels are launched out of profiler step, so don't count them.
                    if not (op.name == 'CallTreeRoot' and rt.name == 'dummy'):
                        for k in rt.get_kernels():
                            sum_time += k.end_time - k.start_time
            return sum_time

        ops_bottom_tc_eligible = []
        for root in tid2tree.values():
            ops_bottom_tc_eligible.extend(TensorCoresParser._get_bottom_tc_eligible_operators(root))
        ops_bottom_tc_eligible_kernel_sum = sum_self_kernel_time(ops_bottom_tc_eligible)
        ops_kernel_sum = sum_self_kernel_time(ops)
        tc_eligible_ops_kernel_ratio = ops_bottom_tc_eligible_kernel_sum / ops_kernel_sum \
            if ops_kernel_sum > 0 else 0.0
        return tc_eligible_ops_kernel_ratio
